@@ -1,7 +1,6 @@
 # coding=utf-8
 from __future__ import unicode_literals
 
-from collections import defaultdict
 from locale import getlocale
 from logging import getLogger
 
@@ -9,11 +8,10 @@ from cached_property import threaded_cached_property
 from future.utils import python_2_unicode_compatible
 from six import string_types
 
-from exchangelib.services import GetUserOofSettings, SetUserOofSettings
-from exchangelib.settings import OofSettings
 from .autodiscover import discover
+from .configuration import Configuration
 from .credentials import DELEGATE, IMPERSONATION, ACCESS_TYPES
-from .errors import ErrorAccessDenied, UnknownTimeZone
+from .errors import UnknownTimeZone
 from .ewsdatetime import EWSTimeZone, UTC
 from .fields import FieldPath
 from .folders import Folder, AdminAuditLogs, ArchiveDeletedItems, ArchiveInbox, ArchiveMsgFolderRoot, \
@@ -22,17 +20,17 @@ from .folders import Folder, AdminAuditLogs, ArchiveDeletedItems, ArchiveInbox,
     Directory, Drafts, Favorites, IMContactList, Inbox, Journal, JunkEmail, LocalFailures, MsgFolderRoot, MyContacts, \
     Notes, Outbox, PeopleConnect, PublicFoldersRoot, QuickContacts, RecipientCache, RecoverableItemsDeletions, \
     RecoverableItemsPurges, RecoverableItemsRoot, RecoverableItemsVersions, Root, SearchFolders, SentItems, \
-    ServerFailures, SyncIssues, Tasks, ToDoSearch, VoiceMail
+    ServerFailures, SyncIssues, Tasks, ToDoSearch, VoiceMail, BaseFolder
 from .items import Item, BulkCreateResult, HARD_DELETE, \
     AUTO_RESOLVE, SEND_TO_NONE, SAVE_ONLY, SEND_AND_SAVE_COPY, SEND_ONLY, ALL_OCCURRENCIES, \
     DELETE_TYPE_CHOICES, MESSAGE_DISPOSITION_CHOICES, CONFLICT_RESOLUTION_CHOICES, AFFECTED_TASK_OCCURRENCES_CHOICES, \
     SEND_MEETING_INVITATIONS_CHOICES, SEND_MEETING_INVITATIONS_AND_CANCELLATIONS_CHOICES, \
-    SEND_MEETING_CANCELLATIONS_CHOICES, IdOnly
-from .properties import Mailbox
-from .protocol import Protocol
+    SEND_MEETING_CANCELLATIONS_CHOICES, ID_ONLY
+from .properties import Mailbox, SendingAs, FolderId, DistinguishedFolderId
 from .queryset import QuerySet
 from .services import ExportItems, UploadItems, GetItem, CreateItem, UpdateItem, DeleteItem, MoveItem, SendItem, \
-    CopyItem
+    CopyItem, GetUserOofSettings, SetUserOofSettings, GetMailTips, ArchiveItem, GetDelegate
+from .settings import OofSettings
 from .util import get_domain, peek
 
 log = getLogger(__name__)
@@ -57,33 +55,21 @@ class Account(object):
         assume values to be in the provided timezone. Defaults to the timezone of the host.
         """
         if '@' not in primary_smtp_address:
-            raise ValueError("primary_smtp_address '%s' is not an email address" % primary_smtp_address)
+            raise ValueError("primary_smtp_address %r is not an email address" % primary_smtp_address)
         self.primary_smtp_address = primary_smtp_address
         self.fullname = fullname
+        # Assume delegate access if individual credentials are provided. Else, assume service user with impersonation
+        self.access_type = access_type or (DELEGATE if credentials else IMPERSONATION)
+        if self.access_type not in ACCESS_TYPES:
+            raise ValueError("'access_type' %r must be one of %s" % (self.access_type, ACCESS_TYPES))
         try:
             self.locale = locale or getlocale()[0] or None  # get_locale() might not be able to determine the locale
         except ValueError as e:
             # getlocale() may throw ValueError if it fails to parse the system locale
             log.warning('Failed to get locale (%s)' % e)
             self.locale = None
-        if self.locale is not None:
-            if not isinstance(self.locale, string_types):
-                raise ValueError("Expected 'locale' to be a string, got %s" % self.locale)
-        # Assume delegate access if individual credentials are provided. Else, assume service user with impersonation
-        self.access_type = access_type or (DELEGATE if credentials else IMPERSONATION)
-        if self.access_type not in ACCESS_TYPES:
-            raise ValueError("'access_type' %s must be one of %s" % (self.access_type, ACCESS_TYPES))
-        if autodiscover:
-            if not credentials:
-                raise AttributeError('autodiscover requires credentials')
-            if config:
-                raise AttributeError('config is ignored when autodiscover is active')
-            self.primary_smtp_address, self.protocol = discover(email=self.primary_smtp_address,
-                                                                credentials=credentials)
-        else:
-            if not config:
-                raise AttributeError('non-autodiscover requires a config')
-            self.protocol = config.protocol
+        if not isinstance(self.locale, string_types + (type(None),)):
+            raise ValueError("Expected 'locale' to be a string, got %r" % self.locale)
         try:
             self.default_timezone = default_timezone or EWSTimeZone.localzone()
         except (ValueError, UnknownTimeZone) as e:
@@ -92,72 +78,70 @@ class Account(object):
             log.warning('%s. Fallback to UTC', e.args[0])
             self.default_timezone = UTC
         if not isinstance(self.default_timezone, EWSTimeZone):
-            raise ValueError("Expected 'default_timezone' to be an EWSTimeZone, got %s" % self.default_timezone)
+            raise ValueError("Expected 'default_timezone' to be an EWSTimeZone, got %r" % self.default_timezone)
+        if not isinstance(config, (Configuration, type(None))):
+            raise ValueError("Expected 'config' to be a Configuration, got %r" % config)
+        if autodiscover:
+            if config:
+                retry_policy, auth_type = config.retry_policy, config.auth_type
+                if not credentials:
+                    credentials = config.credentials
+            else:
+                retry_policy, auth_type = None, None
+            self.primary_smtp_address, self.protocol = discover(
+                email=self.primary_smtp_address, credentials=credentials, auth_type=auth_type, retry_policy=retry_policy
+            )
+        else:
+            if not config:
+                raise AttributeError('non-autodiscover requires a config')
+            self.protocol = config.protocol
         # We may need to override the default server version on a per-account basis because Microsoft may report one
         # server version up-front but delegate account requests to an older backend server.
         self.version = self.protocol.version
-        try:
-            self.root = Root.get_distinguished(account=self)
-        except ErrorAccessDenied:
-            # We may not have access to folder services. This will leave the account severely crippled, but at least
-            # survive the error.
-            log.warning('Access denied to root folder')
-            self.root = Root(account=self)
-
-        if not isinstance(self.protocol, Protocol):
-            raise ValueError("Expected 'protocol' to be a Protocol, got %s" % self.protocol)
         log.debug('Added account: %s', self)
 
-    @property
-    def folders(self):
-        import warnings
-        warnings.warn('The Account.folders mapping is deprecated. Use Account.root.walk() instead')
-        folders_map = defaultdict(list)
-        for f in self.root.walk():
-            folders_map[f.__class__].append(f)
-        return folders_map
-
     @threaded_cached_property
     def admin_audit_logs(self):
         return self.root.get_default_folder(AdminAuditLogs)
 
     @threaded_cached_property
     def archive_deleted_items(self):
-        return self.root.get_default_folder(ArchiveDeletedItems)
+        return self.archive_root.get_default_folder(ArchiveDeletedItems)
 
     @threaded_cached_property
     def archive_inbox(self):
-        return self.root.get_default_folder(ArchiveInbox)
+        return self.archive_root.get_default_folder(ArchiveInbox)
 
     @threaded_cached_property
     def archive_msg_folder_root(self):
-        return self.root.get_default_folder(ArchiveMsgFolderRoot)
+        return self.archive_root.get_default_folder(ArchiveMsgFolderRoot)
 
     @threaded_cached_property
     def archive_recoverable_items_deletions(self):
-        return self.root.get_default_folder(ArchiveRecoverableItemsDeletions)
+        return self.archive_root.get_default_folder(ArchiveRecoverableItemsDeletions)
 
     @threaded_cached_property
     def archive_recoverable_items_purges(self):
-        return self.root.get_default_folder(ArchiveRecoverableItemsPurges)
+        return self.archive_root.get_default_folder(ArchiveRecoverableItemsPurges)
 
     @threaded_cached_property
     def archive_recoverable_items_root(self):
-        return self.root.get_default_folder(ArchiveRecoverableItemsRoot)
+        return self.archive_root.get_default_folder(ArchiveRecoverableItemsRoot)
 
     @threaded_cached_property
     def archive_recoverable_items_versions(self):
-        return self.root.get_default_folder(ArchiveRecoverableItemsVersions)
+        return self.archive_root.get_default_folder(ArchiveRecoverableItemsVersions)
 
     @threaded_cached_property
     def archive_root(self):
-        return self.root.get_default_folder(ArchiveRoot)
+        return ArchiveRoot.get_distinguished(account=self)
 
     @threaded_cached_property
     def calendar(self):
         # If the account contains a shared calendar from a different user, that calendar will be in the folder list.
         # Attempt not to return one of those. An account may not always have a calendar called "Calendar", but a
-        # Calendar folder with a localized name instead. Return that, if it's available.
+        # Calendar folder with a localized name instead. Return that, if it's available, but always prefer any
+        # distinguished folder returned by the server.
         return self.root.get_default_folder(Calendar)
 
     @threaded_cached_property
@@ -226,7 +210,7 @@ class Account(object):
 
     @threaded_cached_property
     def public_folders_root(self):
-        return self.root.get_default_folder(PublicFoldersRoot)
+        return PublicFoldersRoot.get_distinguished(account=self)
 
     @threaded_cached_property
     def quick_contacts(self):
@@ -252,6 +236,10 @@ class Account(object):
     def recoverable_items_versions(self):
         return self.root.get_default_folder(RecoverableItemsVersions)
 
+    @threaded_cached_property
+    def root(self):
+        return Root.get_distinguished(account=self)
+
     @threaded_cached_property
     def search_folders(self):
         return self.root.get_default_folder(SearchFolders)
@@ -307,6 +295,21 @@ class Account(object):
             oof_settings=value,
         )
 
+    def _consume_item_service(self, service_cls, items, chunk_size, kwargs):
+        # 'items' could be an unevaluated QuerySet, e.g. if we ended up here via `some_folder.filter(...).delete()`. In
+        # that case, we want to use its iterator. Otherwise, peek() will start a count() which is wasteful because we
+        # need the item IDs immediately afterwards. iterator() will only do the bare minimum.
+        if isinstance(items, QuerySet):
+            items = items.iterator()
+        is_empty, items = peek(items)
+        if is_empty:
+            # We accept generators, so it's not always convenient for caller to know up-front if 'ids' is empty. Allow
+            # empty 'ids' and return early.
+            return
+        kwargs['items'] = items
+        for i in service_cls(account=self, chunk_size=chunk_size).call(**kwargs):
+            yield i
+
     def export(self, items, chunk_size=None):
         """Return export strings of the given items
 
@@ -315,12 +318,9 @@ class Account(object):
 
         :return A list of strings, the exported representation of the object
         """
-        is_empty, items = peek(items)
-        if is_empty:
-            # We accept generators, so it's not always convenient for caller to know up-front if 'items' is empty. Allow
-            # empty 'items' and return early.
-            return []
-        return list(ExportItems(account=self, chunk_size=chunk_size).call(items=items))
+        return list(
+            self._consume_item_service(service_cls=ExportItems, items=items, chunk_size=chunk_size, kwargs=dict())
+        )
 
     def upload(self, data, chunk_size=None):
         """Adds objects retrieved from export into the given folders
@@ -356,8 +356,8 @@ class Account(object):
                SEND_MEETING_INVITATIONS_CHOICES
         :param chunk_size: The number of items to send to the server in a single request
         :return: a list of either BulkCreateResult or exception instances in the same order as the input. The returned
-                 BulkCreateResult objects are normal Item objects except they only contain the 'item_id' and 'changekey'
-                 of the created item, and the 'item_id' on any attachments that were also created.
+                 BulkCreateResult objects are normal Item objects except they only contain the 'id' and 'changekey'
+                 of the created item, and the 'id' of any attachments that were also created.
         """
         if message_disposition not in MESSAGE_DISPOSITION_CHOICES:
             raise ValueError("'message_disposition' %s must be one of %s" % (
@@ -368,8 +368,8 @@ class Account(object):
                 send_meeting_invitations, SEND_MEETING_INVITATIONS_CHOICES
             ))
         if folder is not None:
-            if not isinstance(folder, Folder):
-                raise ValueError("'folder' %r must be a Folder instence" % folder)
+            if not isinstance(folder, BaseFolder):
+                raise ValueError("'folder' %r must be a Folder instance" % folder)
             if folder.account != self:
                 raise ValueError('"Folder must belong to this account')
         if message_disposition == SAVE_ONLY and folder is None:
@@ -388,20 +388,14 @@ class Account(object):
             message_disposition,
             send_meeting_invitations,
         )
-        is_empty, items = peek(items)
-        if is_empty:
-            # We accept generators, so it's not always convenient for caller to know up-front if 'items' is empty. Allow
-            # empty 'items' and return early.
-            return []
         return list(
             i if isinstance(i, Exception)
             else BulkCreateResult.from_xml(elem=i, account=self)
-            for i in CreateItem(account=self, chunk_size=chunk_size).call(
-                items=items,
+            for i in self._consume_item_service(service_cls=CreateItem, items=items, chunk_size=chunk_size, kwargs=dict(
                 folder=folder,
                 message_disposition=message_disposition,
                 send_meeting_invitations=send_meeting_invitations,
-            )
+            ))
         )
 
     def bulk_update(self, items, conflict_resolution=AUTO_RESOLVE, message_disposition=SAVE_ONLY,
@@ -420,7 +414,7 @@ class Account(object):
         :param suppress_read_receipts: nly supported from Exchange 2013. True or False
         :param chunk_size: The number of items to send to the server in a single request
 
-        :return: a list of either (item_id, changekey) tuples or exception instances, in the same order as the input
+        :return: a list of either (id, changekey) tuples or exception instances, in the same order as the input
         """
         if conflict_resolution not in CONFLICT_RESOLUTION_CHOICES:
             raise ValueError("'conflict_resolution' %s must be one of %s" % (
@@ -450,20 +444,14 @@ class Account(object):
             message_disposition,
             send_meeting_invitations_or_cancellations,
         )
-        is_empty, items = peek(items)
-        if is_empty:
-            # We accept generators, so it's not always convenient for caller to know up-front if 'items' is empty. Allow
-            # empty 'items' and return early.
-            return []
         return list(
             i if isinstance(i, Exception) else Item.id_from_xml(i)
-            for i in UpdateItem(account=self, chunk_size=chunk_size).call(
-                items=items,
+            for i in self._consume_item_service(service_cls=UpdateItem, items=items, chunk_size=chunk_size, kwargs=dict(
                 conflict_resolution=conflict_resolution,
                 message_disposition=message_disposition,
                 send_meeting_invitations_or_cancellations=send_meeting_invitations_or_cancellations,
                 suppress_read_receipts=suppress_read_receipts,
-            )
+            ))
         )
 
     def bulk_delete(self, ids, delete_type=HARD_DELETE, send_meeting_cancellations=SEND_TO_NONE,
@@ -471,7 +459,7 @@ class Account(object):
         """
         Bulk deletes items.
 
-        :param ids: an iterable of either (item_id, changekey) tuples or Item objects.
+        :param ids: an iterable of either (id, changekey) tuples or Item objects.
         :param delete_type: the type of delete to perform. Possible values are specified in DELETE_TYPE_CHOICES
         :param send_meeting_cancellations: only applicable to CalendarItem. Possible values are specified in
                SEND_MEETING_CANCELLATIONS_CHOICES.
@@ -503,28 +491,19 @@ class Account(object):
             send_meeting_cancellations,
             affected_task_occurrences,
         )
-        # 'ids' could be an unevaluated QuerySet, e.g. if we ended up here via `some_folder.filter(...).delete()`. In
-        # that case, we want to use its iterator. Otherwise, peek() will start a count() which is wasteful because we
-        # need the item IDs immediately afterwards. iterator() will only do the bare minimum.
-        if isinstance(ids, QuerySet):
-            ids = ids.iterator()
-        is_empty, ids = peek(ids)
-        if is_empty:
-            # We accept generators, so it's not always convenient for caller to know up-front if 'ids' is empty. Allow
-            # empty 'ids' and return early.
-            return []
-        return list(DeleteItem(account=self, chunk_size=chunk_size).call(
-            items=ids,
-            delete_type=delete_type,
-            send_meeting_cancellations=send_meeting_cancellations,
-            affected_task_occurrences=affected_task_occurrences,
-            suppress_read_receipts=suppress_read_receipts,
-        ))
+        return list(
+            self._consume_item_service(service_cls=DeleteItem, items=ids, chunk_size=chunk_size, kwargs=dict(
+                delete_type=delete_type,
+                send_meeting_cancellations=send_meeting_cancellations,
+                affected_task_occurrences=affected_task_occurrences,
+                suppress_read_receipts=suppress_read_receipts,
+            ))
+        )
 
     def bulk_send(self, ids, save_copy=True, copy_to_folder=None, chunk_size=None):
         """ Send existing draft messages. If requested, save a copy in 'copy_to_folder'
 
-        :param ids: an iterable of either (item_id, changekey) tuples or Item objects.
+        :param ids: an iterable of either (id, changekey) tuples or Item objects.
         :param save_copy: If true, saves a copy of the message
         :param copy_to_folder: If requested, save a copy of the message in this folder. Default is the Sent folder
         :param chunk_size: The number of items to send to the server in a single request
@@ -534,105 +513,120 @@ class Account(object):
             raise AttributeError("'save_copy' must be True when 'copy_to_folder' is set")
         if save_copy and not copy_to_folder:
             copy_to_folder = self.sent  # 'Sent' is default EWS behaviour
-        # 'ids' could be an unevaluated QuerySet, e.g. if we ended up here via `bulk_send(some_folder.filter(...))`. In
-        # that case, we want to use its iterator. Otherwise, peek() will start a count() which is wasteful because we
-        # need the item IDs immediately afterwards. iterator() will only do the bare minimum.
-        if isinstance(ids, QuerySet):
-            ids = ids.iterator()
-        is_empty, ids = peek(ids)
-        if is_empty:
-            # We accept generators, so it's not always convenient for caller to know up-front if 'ids' is empty. Allow
-            # empty 'ids' and return early.
-            return []
-        return list(SendItem(account=self, chunk_size=chunk_size).call(items=ids, saved_item_folder=copy_to_folder))
+        if copy_to_folder and not isinstance(copy_to_folder, BaseFolder):
+            raise ValueError("'copy_to_folder' %r must be a Folder instance" % copy_to_folder)
+        return list(
+            self._consume_item_service(service_cls=SendItem, items=ids, chunk_size=chunk_size, kwargs=dict(
+                saved_item_folder=copy_to_folder,
+            ))
+        )
 
     def bulk_copy(self, ids, to_folder, chunk_size=None):
         """ Copy items to another folder
 
-        :param ids: an iterable of either (item_id, changekey) tuples or Item objects.
+        :param ids: an iterable of either (id, changekey) tuples or Item objects.
         :param to_folder: The destination folder of the copy operation
         :param chunk_size: The number of items to send to the server in a single request
         :return: Status for each send operation, in the same order as the input
         """
-        if not isinstance(to_folder, Folder):
-            raise ValueError("'to_folder' %r must be a Folder instence" % to_folder)
-        # 'ids' could be an unevaluated QuerySet, e.g. if we ended up here via `bulk_copy(some_folder.filter(...))`. In
-        # that case, we want to use its iterator. Otherwise, peek() will start a count() which is wasteful because we
-        # need the item IDs immediately afterwards. iterator() will only do the bare minimum.
-        if isinstance(ids, QuerySet):
-            ids = ids.iterator()
-        is_empty, ids = peek(ids)
-        if is_empty:
-            # We accept generators, so it's not always convenient for caller to know up-front if 'ids' is empty. Allow
-            # empty 'ids' and return early.
-            return []
+        if not isinstance(to_folder, BaseFolder):
+            raise ValueError("'to_folder' %r must be a Folder instance" % to_folder)
         return list(
             i if isinstance(i, Exception) else Item.id_from_xml(i)
-            for i in CopyItem(account=self, chunk_size=chunk_size).call(items=ids, to_folder=to_folder)
+            for i in self._consume_item_service(service_cls=CopyItem, items=ids, chunk_size=chunk_size, kwargs=dict(
+                to_folder=to_folder,
+            ))
         )
 
     def bulk_move(self, ids, to_folder, chunk_size=None):
         """Move items to another folder
 
-        :param ids: an iterable of either (item_id, changekey) tuples or Item objects.
+        :param ids: an iterable of either (id, changekey) tuples or Item objects.
         :param to_folder: The destination folder of the copy operation
         :param chunk_size: The number of items to send to the server in a single request
-        :return: The new IDs of the moved items, in the same order as the input
+        :return: The new IDs of the moved items, in the same order as the input. If 'to_folder' is a public folder or a
+        folder in a different mailbox, an empty list is returned.
         """
-        if not isinstance(to_folder, Folder):
-            raise ValueError("'to_folder' %r must be a Folder instence" % to_folder)
-        # 'ids' could be an unevaluated QuerySet, e.g. if we ended up here via `bulk_move(some_folder.filter(...))`. In
-        # that case, we want to use its iterator. Otherwise, peek() will start a count() which is wasteful because we
-        # need the item IDs immediately afterwards. iterator() will only do the bare minimum.
-        if isinstance(ids, QuerySet):
-            ids = ids.iterator()
-        is_empty, ids = peek(ids)
-        if is_empty:
-            # We accept generators, so it's not always convenient for caller to know up-front if 'ids' is empty. Allow
-            # empty 'ids' and return early.
-            return []
+        if not isinstance(to_folder, BaseFolder):
+            raise ValueError("'to_folder' %r must be a Folder instance" % to_folder)
         return list(
             i if isinstance(i, Exception) else Item.id_from_xml(i)
-            for i in MoveItem(account=self, chunk_size=chunk_size).call(items=ids, to_folder=to_folder)
+            for i in self._consume_item_service(service_cls=MoveItem, items=ids, chunk_size=chunk_size, kwargs=dict(
+                to_folder=to_folder,
+            ))
+        )
+
+    def bulk_archive(self, ids, to_folder, chunk_size=None):
+        """Archive items to a folder in the archive mailbox. An archive mailbox must be enabled in order for this
+        to work.
+
+        :param ids: an iterable of either (id, changekey) tuples or Item objects.
+        :param to_folder: The destination folder of the archive operation
+        :param chunk_size: The number of items to send to the server in a single request
+        :return: A list containing True or an exception instance in stable order of the requested items
+        """
+        if not isinstance(to_folder, (BaseFolder, FolderId, DistinguishedFolderId)):
+            raise ValueError("'to_folder' %r must be a Folder or FolderId instance" % to_folder)
+        return list(self._consume_item_service(service_cls=ArchiveItem, items=ids, chunk_size=chunk_size, kwargs=dict(
+                to_folder=to_folder,
+            ))
         )
 
     def fetch(self, ids, folder=None, only_fields=None, chunk_size=None):
         """ Fetch items by ID
 
-        :param ids: an iterable of either (item_id, changekey) tuples or Item objects.
+        :param ids: an iterable of either (id, changekey) tuples or Item objects.
         :param folder: used for validating 'only_fields'
         :param only_fields: A list of string or FieldPath items specifying the fields to fetch. Default to all fields
         :param chunk_size: The number of items to send to the server in a single request
         :return: A generator of Item objects, in the same order as the input
         """
-        validation_folder = folder or Folder(account=self)  # Default to a folder type that supports all item types
+        validation_folder = folder or Folder(root=self.root)  # Default to a folder type that supports all item types
         # 'ids' could be an unevaluated QuerySet, e.g. if we ended up here via `fetch(ids=some_folder.filter(...))`. In
         # that case, we want to use its iterator. Otherwise, peek() will start a count() which is wasteful because we
         # need the item IDs immediately afterwards. iterator() will only do the bare minimum.
-        if isinstance(ids, QuerySet):
-            ids = ids.iterator()
-        is_empty, ids = peek(ids)
-        if is_empty:
-            # We accept generators, so it's not always convenient for caller to know up-front if 'ids' is empty. Allow
-            # empty 'ids' and return early.
-            return
         if only_fields is None:
             # We didn't restrict list of field paths. Get all fields from the server, including extended properties.
-            additional_fields = {FieldPath(field=f) for f in validation_folder.allowed_fields()}
+            additional_fields = {
+                FieldPath(field=f) for f in validation_folder.allowed_item_fields(version=self.version)
+            }
         else:
-            additional_fields = validation_folder.validate_fields(fields=only_fields)
+            for field in only_fields:
+                validation_folder.validate_item_field(field=field)
+            additional_fields = validation_folder.normalize_fields(fields=only_fields)
         # Always use IdOnly here, because AllProperties doesn't actually get *all* properties
-        for i in GetItem(account=self, chunk_size=chunk_size).call(
-                items=ids,
+        for i in self._consume_item_service(service_cls=GetItem, items=ids, chunk_size=chunk_size, kwargs=dict(
                 additional_fields=additional_fields,
-                shape=IdOnly,
-        ):
+                shape=ID_ONLY,
+        )):
             if isinstance(i, Exception):
                 yield i
             else:
                 item = validation_folder.item_model_from_tag(i.tag).from_xml(elem=i, account=self)
                 yield item
 
+    @property
+    def mail_tips(self):
+        """See self.oof_settings about caching considerations
+        """
+        # mail_tips_requested must be one of properties.MAIL_TIPS_TYPES
+        res = list(GetMailTips(protocol=self.protocol).call(
+            sending_as=SendingAs(email_address=self.primary_smtp_address),
+            recipients=[Mailbox(email_address=self.primary_smtp_address)],
+            mail_tips_requested='All',
+        ))
+        if len(res) != 1:
+            raise ValueError('Expected result length 1, but got %s' % res)
+        if isinstance(res[0], Exception):
+            raise res[0]
+        return res[0]
+
+    @property
+    def delegates(self):
+        """Returns a list of DelegateUser objects representing the delegates that are set on this account
+        """
+        return list(GetDelegate(account=self).call(user_ids=None, include_permissions=True))
+
     def __str__(self):
         txt = '%s' % self.primary_smtp_address
         if self.fullname:
