@@ -1,6 +1,7 @@
 # coding=utf-8
+# flake8: noqa
 """
-Stores errors specific to exchangelib, and mirrors all the possible errors that EWS can return.
+Stores errors specific to this package, and mirrors all the possible errors that EWS can return.
 """
 from __future__ import unicode_literals
 
@@ -8,14 +9,21 @@ from future.moves.urllib.parse import urlparse
 from future.utils import python_2_unicode_compatible
 from six import text_type
 
+import pytz.exceptions
+
+
+class MultipleObjectsReturned(Exception):
+    pass
+
+
+class DoesNotExist(Exception):
+    pass
+
 
 @python_2_unicode_compatible
 class EWSError(Exception):
+    """Global error type within this module.
     """
-    Global error type within this module.
-
-    """
-
     def __init__(self, value):
         super(EWSError, self).__init__(value)
         self.value = value
@@ -35,13 +43,27 @@ class TransportError(EWSError):
 
 
 class RateLimitError(TransportError):
-    pass
+    def __init__(self, value, url, status_code, total_wait):
+        super(RateLimitError, self).__init__(value)
+        self.url = url
+        self.status_code = status_code
+        self.total_wait = total_wait
+
+    def __str__(self):
+        return text_type(
+            '{value} (gave up after {total_wait} seconds. URL {url} returned status code {status_code})'.format(
+                value=self.value, url=self.url, status_code=self.status_code, total_wait=self.total_wait)
+        )
 
 
 class SOAPError(TransportError):
     pass
 
 
+class MalformedResponseError(TransportError):
+    pass
+
+
 class UnauthorizedError(EWSError):
     pass
 
@@ -93,11 +115,15 @@ class UnknownTimeZone(EWSError):
     pass
 
 
-class AmbiguousTimeError(EWSError):
+class AmbiguousTimeError(EWSError, pytz.exceptions.AmbiguousTimeError):
     pass
 
 
-class NonExistentTimeError(EWSError):
+class NonExistentTimeError(EWSError, pytz.exceptions.NonExistentTimeError):
+    pass
+
+
+class SessionPoolMinSizeReached(EWSError):
     pass
 
 
@@ -107,8 +133,9 @@ class ResponseMessageError(TransportError):
 
 @python_2_unicode_compatible
 class CASError(EWSError):
-    # EWS will sometimes return an error message in an 'X-CasErrorCode' custom HTTP header in an HTTP 500 error code.
-    # This exception is for those cases. The caller may want to do something with the original response, so store that.
+    """EWS will sometimes return an error message in an 'X-CasErrorCode' custom HTTP header in an HTTP 500 error code.
+    This exception is for those cases. The caller may want to do something with the original response, so store that.
+    """
     def __init__(self, cas_error, response):
         self.cas_error = cas_error
         self.response = response
@@ -119,7 +146,7 @@ class CASError(EWSError):
 
 
 # Somewhat-authoritative list of possible response message error types from EWS. See full list at
-# https://msdn.microsoft.com/en-us/library/office/aa580757(v=exchg.150).aspx
+# https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/responsecode
 #
 class ErrorAccessDenied(ResponseMessageError): pass
 class ErrorAccessModeSpecified(ResponseMessageError): pass
