@@ -3,6 +3,7 @@ from __future__ import unicode_literals
 import abc
 import base64
 import binascii
+from collections import OrderedDict
 import datetime
 from decimal import Decimal, InvalidOperation
 import logging
@@ -10,16 +11,15 @@ import logging
 from six import string_types
 
 from .errors import ErrorInvalidServerVersion
-from .ewsdatetime import EWSDateTime, EWSDate, EWSTimeZone, NaiveDateTimeNotAllowed, UnknownTimeZone, UTC
-from .services import TNS
-from .util import create_element, get_xml_attrs, set_xml_value, value_to_xml_text, is_iterable
+from .ewsdatetime import EWSDateTime, EWSDate, EWSTimeZone, NaiveDateTimeNotAllowed, UnknownTimeZone
+from .util import create_element, get_xml_attrs, set_xml_value, value_to_xml_text, is_iterable, TNS
 from .version import Build, EXCHANGE_2013
 
-string_type = string_types[0]
 log = logging.getLogger(__name__)
 
 
-# DayOfWeekIndex enum. See https://msdn.microsoft.com/en-us/library/office/aa581350(v=exchg.150).aspx
+# DayOfWeekIndex enum. See
+# https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/dayofweekindex
 FIRST = 'First'
 SECOND = 'Second'
 THIRD = 'Third'
@@ -58,7 +58,8 @@ WEEK_DAY = 'Weekday'  # Non-weekend day
 WEEKEND_DAY = 'WeekendDay'
 EXTRA_WEEKDAY_OPTIONS = (DAY, WEEK_DAY, WEEKEND_DAY)
 
-# DaysOfWeek enum: See https://msdn.microsoft.com/en-us/library/office/ee332417(v=exchg.150).aspx
+# DaysOfWeek enum: See
+# https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/daysofweek-daysofweektype
 WEEKDAYS = WEEKDAY_NAMES + EXTRA_WEEKDAY_OPTIONS
 
 
@@ -97,7 +98,9 @@ def resolve_field_path(field_path, folder, strict=True):
                 "IndexedField path '%s' must specify label, e.g. '%s__%s'"
                 % (field_path, fieldname, field.value_cls.get_field_by_fieldname('label').default)
             )
-        valid_labels = field.value_cls.get_field_by_fieldname('label').supported_choices(version=folder.account.version)
+        valid_labels = field.value_cls.get_field_by_fieldname('label').supported_choices(
+            version=folder.account.version
+        )
         if label and label not in valid_labels:
             raise ValueError(
                 "Label '%s' on IndexedField path '%s' must be one of %s"
@@ -114,7 +117,9 @@ def resolve_field_path(field_path, folder, strict=True):
                 try:
                     subfield = field.value_cls.get_field_by_fieldname(subfieldname)
                 except ValueError:
-                    fnames = ', '.join(f.name for f in field.value_cls.supported_fields(version=folder.account.version))
+                    fnames = ', '.join(f.name for f in field.value_cls.supported_fields(
+                        version=folder.account.version
+                    ))
                     raise ValueError(
                         "Subfield '%s' on IndexedField path '%s' must be one of %s"
                         % (subfieldname, field_path, fnames)
@@ -138,7 +143,7 @@ def resolve_field_path(field_path, folder, strict=True):
 
 
 class FieldPath(object):
-    """ Holds values needed to point to a single field. For indexed properties, we allow setting eiterh field,
+    """ Holds values needed to point to a single field. For indexed properties, we allow setting either field,
     field and label, or field, label and subfield. This allows pointing to either the full indexed property set, a
     property with a specific label, or a particular subfield field on that property. """
     def __init__(self, field, label=None, subfield=None):
@@ -154,8 +159,8 @@ class FieldPath(object):
         self.subfield = subfield
 
     @classmethod
-    def from_string(cls, s, folder, strict=False):
-        field, label, subfield = resolve_field_path(s, folder=folder, strict=strict)
+    def from_string(cls, field_path, folder, strict=False):
+        field, label, subfield = resolve_field_path(field_path, folder=folder, strict=strict)
         return cls(field=field, label=label, subfield=subfield)
 
     def get_value(self, item):
@@ -203,6 +208,12 @@ class FieldPath(object):
     def __eq__(self, other):
         return hash(self) == hash(other)
 
+    def __str__(self):
+        return self.path
+
+    def __repr__(self):
+        return self.__class__.__name__ + repr((self.field, self.label, self.subfield))
+
     def __hash__(self):
         return hash((self.field, self.label, self.subfield))
 
@@ -210,17 +221,22 @@ class FieldPath(object):
 class FieldOrder(object):
     """ Holds values needed to call server-side sorting on a single field path """
     def __init__(self, field_path, reverse=False):
+        if not isinstance(field_path, FieldPath):
+            raise ValueError("'field_path' %r must be a FieldPath instance" % field_path)
+        if not isinstance(reverse, bool):
+            raise ValueError("'reverse' %r must be a boolean" % reverse)
         self.field_path = field_path
         self.reverse = reverse
 
     @classmethod
-    def from_string(cls, s, folder):
-        field_path = FieldPath.from_string(s.lstrip('-'), folder=folder, strict=True)
-        reverse = s.startswith('-')
-        return cls(field_path=field_path, reverse=reverse)
+    def from_string(cls, field_path, folder):
+        return cls(
+            field_path=FieldPath.from_string(field_path=field_path.lstrip('-'), folder=folder, strict=True),
+            reverse=field_path.startswith('-')
+        )
 
     def to_xml(self):
-        field_order = create_element('t:FieldOrder', Order='Descending' if self.reverse else 'Ascending')
+        field_order = create_element('t:FieldOrder', attrs=dict(Order='Descending' if self.reverse else 'Ascending'))
         field_order.append(self.field_path.to_xml())
         return field_order
 
@@ -281,7 +297,7 @@ class Field(object):
                 raise ValueError("Field '%s' value %r must be a list" % (self.name, value))
             for v in value:
                 if not isinstance(v, self.value_cls):
-                    raise TypeError('Field %s value "%r must be of type %s' % (self.name, v, self.value_cls))
+                    raise TypeError("Field '%s' value %r must be of type %s" % (self.name, v, self.value_cls))
                 if hasattr(v, 'clean'):
                     v.clean(version=version)
         else:
@@ -322,10 +338,13 @@ class Field(object):
 
 
 class FieldURIField(Field):
+    namespace = TNS
+
     def __init__(self, *args, **kwargs):
         self.field_uri = kwargs.pop('field_uri', None)
         super(FieldURIField, self).__init__(*args, **kwargs)
-        # See all valid FieldURI values at https://msdn.microsoft.com/en-us/library/office/aa494315(v=exchg.150).aspx
+        # See all valid FieldURI values at
+        # https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/fielduri
         # The field_uri has a prefix when the FieldURI points to an Item field.
         if self.field_uri is None:
             self.field_uri_postfix = None
@@ -334,6 +353,12 @@ class FieldURIField(Field):
         else:
             self.field_uri_postfix = self.field_uri
 
+    def _get_val_from_elem(self, elem):
+        if self.is_attribute:
+            return elem.get(self.field_uri)
+        field_elem = elem.find(self.response_tag())
+        return None if field_elem is None else field_elem.text or None
+
     def from_xml(self, elem, account):
         raise NotImplementedError()
 
@@ -344,7 +369,7 @@ class FieldURIField(Field):
     def field_uri_xml(self):
         if not self.field_uri:
             raise ValueError("'field_uri' value is missing")
-        return create_element('t:FieldURI', FieldURI=self.field_uri)
+        return create_element('t:FieldURI', attrs=dict(FieldURI=self.field_uri))
 
     def request_tag(self):
         if not self.field_uri_postfix:
@@ -354,7 +379,7 @@ class FieldURIField(Field):
     def response_tag(self):
         if not self.field_uri_postfix:
             raise ValueError("'field_uri_postfix' value is missing")
-        return '{%s}%s' % (TNS, self.field_uri_postfix)
+        return '{%s}%s' % (self.namespace, self.field_uri_postfix)
 
     def __hash__(self):
         return hash(self.field_uri)
@@ -364,8 +389,7 @@ class BooleanField(FieldURIField):
     value_cls = bool
 
     def from_xml(self, elem, account):
-        field_elem = elem.find(self.response_tag())
-        val = None if field_elem is None else field_elem.text or None
+        val = self._get_val_from_elem(elem)
         if val is not None:
             try:
                 return {
@@ -393,19 +417,18 @@ class IntegerField(FieldURIField):
                 for v in value:
                     if self.min is not None and v < self.min:
                         raise ValueError(
-                            "value '%s' on field '%s' must be greater than %s" % (value, self.name, self.min))
+                            "Value %r on field '%s' must be greater than %s" % (v, self.name, self.min))
                     if self.max is not None and v > self.max:
-                        raise ValueError("value '%s' on field '%s' must be less than %s" % (value, self.name, self.max))
+                        raise ValueError("Value %r on field '%s' must be less than %s" % (v, self.name, self.max))
             else:
                 if self.min is not None and value < self.min:
-                    raise ValueError("value '%s' on field '%s' must be greater than %s" % (value, self.name, self.min))
+                    raise ValueError("Value %r on field '%s' must be greater than %s" % (value, self.name, self.min))
                 if self.max is not None and value > self.max:
-                    raise ValueError("value '%s' on field '%s' must be less than %s" % (value, self.name, self.max))
+                    raise ValueError("Value %r on field '%s' must be less than %s" % (value, self.name, self.max))
         return value
 
     def from_xml(self, elem, account):
-        field_elem = elem.find(self.response_tag())
-        val = None if field_elem is None else field_elem.text or None
+        val = self._get_val_from_elem(elem)
         if val is not None:
             try:
                 return self.value_cls(val)
@@ -420,8 +443,9 @@ class DecimalField(IntegerField):
 
 
 class EnumField(IntegerField):
-    # A field type where you can enter either the 1-based index in an enum (tuple), or the enum value. Values will be
-    # stored internally as integers but output in XML as strings.
+    """A field type where you can enter either the 1-based index in an enum (tuple), or the enum value. Values will be
+    stored internally as integers but output in XML as strings.
+    """
     def __init__(self, *args, **kwargs):
         self.enum = kwargs.pop('enum')
         # Set different min/max defaults than IntegerField
@@ -452,9 +476,16 @@ class EnumField(IntegerField):
                 value = self.enum.index(value) + 1
         return super(EnumField, self).clean(value, version=version)
 
+    def as_string(self, value):
+        # Converts an integer in the enum to its equivalent string
+        if isinstance(value, string_types):
+            return value
+        if self.is_list:
+            return [self.enum[v - 1] for v in sorted(value)]
+        return self.enum[value - 1]
+
     def from_xml(self, elem, account):
-        field_elem = elem.find(self.response_tag())
-        val = None if field_elem is None else field_elem.text or None
+        val = self._get_val_from_elem(elem)
         if val is not None:
             try:
                 if self.is_list:
@@ -468,8 +499,8 @@ class EnumField(IntegerField):
     def to_xml(self, value, version):
         field_elem = create_element(self.request_tag())
         if self.is_list:
-            return set_xml_value(field_elem, ' '.join(self.enum[v - 1] for v in sorted(value)), version=version)
-        return set_xml_value(field_elem, self.enum[value - 1], version=version)
+            return set_xml_value(field_elem, ' '.join(self.as_string(value)), version=version)
+        return set_xml_value(field_elem, self.as_string(value), version=version)
 
 
 class EnumListField(EnumField):
@@ -477,10 +508,9 @@ class EnumListField(EnumField):
 
 
 class EnumAsIntField(EnumField):
-    # Like EnumField, but communicates values with EWS in integers
+    """Like EnumField, but communicates values with EWS in integers"""
     def from_xml(self, elem, account):
-        field_elem = elem.find(self.response_tag())
-        val = None if field_elem is None else field_elem.text or None
+        val = self._get_val_from_elem(elem)
         if val is not None:
             try:
                 return int(val)
@@ -504,8 +534,7 @@ class Base64Field(FieldURIField):
         super(Base64Field, self).__init__(*args, **kwargs)
 
     def from_xml(self, elem, account):
-        field_elem = elem.find(self.response_tag())
-        val = None if field_elem is None else field_elem.text or None
+        val = self._get_val_from_elem(elem)
         if val is not None:
             try:
                 return base64.b64decode(val)
@@ -519,12 +548,33 @@ class Base64Field(FieldURIField):
         return set_xml_value(field_elem, base64.b64encode(value).decode('ascii'), version=version)
 
 
+class MimeContentField(Base64Field):
+    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/mimecontent"""
+    value_cls = string_types[0]
+    default_encoding = 'utf-8'
+
+    def from_xml(self, elem, account):
+        val = super(MimeContentField, self).from_xml(elem=elem, account=account)
+        if val is None or val == self.default:
+            # Return default values unaltered
+            return val
+
+        charset = elem.find(self.response_tag()).get('CharacterSet').lower() or self.default_encoding
+        try:
+            return val.decode(charset)
+        except UnicodeDecodeError:
+            log.warning("Cannot decode value '%s' on field '%s' to charset %s", val, self.name, self.value_cls)
+            return None
+
+    def to_xml(self, value, version):
+        return super(MimeContentField, self).to_xml(value=value.encode(self.default_encoding), version=version)
+
+
 class DateField(FieldURIField):
     value_cls = EWSDate
 
     def from_xml(self, elem, account):
-        field_elem = elem.find(self.response_tag())
-        val = None if field_elem is None else field_elem.text or None
+        val = self._get_val_from_elem(elem)
         if val is not None:
             try:
                 return self.value_cls.from_string(val)
@@ -538,12 +588,15 @@ class TimeField(FieldURIField):
     value_cls = datetime.time
 
     def from_xml(self, elem, account):
-        field_elem = elem.find(self.response_tag())
-        val = None if field_elem is None else field_elem.text or None
+        val = self._get_val_from_elem(elem)
         if val is not None:
             try:
-                # Assume an integer in minutes since midnight
-                return (datetime.datetime(2000, 1, 1) + datetime.timedelta(minutes=int(val))).time()
+                if ':' in val:
+                    # Assume a string of the form HH:MM:SS
+                    return datetime.datetime.strptime(val, '%H:%M:%S').time()
+                else:
+                    # Assume an integer in minutes since midnight
+                    return (datetime.datetime(2000, 1, 1) + datetime.timedelta(minutes=int(val))).time()
             except ValueError:
                 pass
         return self.default
@@ -558,20 +611,23 @@ class DateTimeField(FieldURIField):
         return super(DateTimeField, self).clean(value, version=version)
 
     def from_xml(self, elem, account):
-        field_elem = elem.find(self.response_tag())
-        val = None if field_elem is None else field_elem.text or None
+        val = self._get_val_from_elem(elem)
         if val is not None:
             try:
                 return self.value_cls.from_string(val)
             except ValueError as e:
                 if isinstance(e, NaiveDateTimeNotAllowed):
-                    # We encountered a naive datetime. Convert to timezone-aware datetime using the default timezone of
-                    # the account.
+                    # We encountered a naive datetime
                     local_dt = e.args[0]
-                    tz = account.default_timezone if account else UTC
-                    log.info('Found naive datetime %s on field %s. Assuming timezone %s', local_dt, self.name, tz)
-                    return tz.localize(local_dt)
-                log.warning("Cannot convert value '%s' on field '%s' to type %s", val, self.name, self.value_cls)
+                    if account:
+                        # Convert to timezone-aware datetime using the default timezone of the account
+                        tz = account.default_timezone
+                        log.info('Found naive datetime %s on field %s. Assuming timezone %s', local_dt, self.name, tz)
+                        return tz.localize(local_dt)
+                    # There's nothing we can do but return the naive date. It's better than assuming e.g. UTC.
+                    log.warning('Returning naive datetime %s on field %s', local_dt, self.name)
+                    return local_dt
+                log.info("Cannot convert value '%s' on field '%s' to type %s", val, self.name, self.value_cls)
                 return None
         return self.default
 
@@ -595,20 +651,25 @@ class TimeZoneField(FieldURIField):
         return self.default
 
     def to_xml(self, value, version):
-        return create_element('t:%s' % self.field_uri_postfix, Id=value.ms_id, Name=value.ms_name)
+        return create_element(
+            't:%s' % self.field_uri_postfix,
+            attrs=OrderedDict([
+                ('Id', value.ms_id),
+                ('Name', value.ms_name),
+            ])
+        )
 
 
 class TextField(FieldURIField):
-    # A field that stores a string value with no length limit
-    value_cls = string_type
+    """A field that stores a string value with no length limit"""
+    value_cls = string_types[0]
     is_complex = True
 
     def from_xml(self, elem, account):
         if self.is_attribute:
             val = elem.get(self.field_uri)
         else:
-            field_elem = elem.find(self.response_tag())
-            val = None if field_elem is None else field_elem.text or None
+            val = self._get_val_from_elem(elem)
         if val is not None:
             return val
         return self.default
@@ -624,8 +685,27 @@ class TextListField(TextField):
         return self.default
 
 
+class MessageField(TextField):
+    INNER_ELEMENT_NAME = 'Message'
+
+    def from_xml(self, elem, account):
+        reply = elem.find(self.response_tag())
+        if reply is None:
+            return None
+        message = reply.find('{%s}%s' % (TNS, self.INNER_ELEMENT_NAME))
+        if message is None:
+            return None
+        return message.text
+
+    def to_xml(self, value, version):
+        field_elem = create_element(self.request_tag())
+        message = create_element('t:%s' % self.INNER_ELEMENT_NAME)
+        message.text = value
+        return set_xml_value(field_elem, message, version=version)
+
+
 class CharField(TextField):
-    # A field that stores a string value with a limited length
+    """A field that stores a string value with a limited length"""
     is_complex = False
 
     def __init__(self, *args, **kwargs):
@@ -649,8 +729,10 @@ class CharField(TextField):
 
 
 class IdField(CharField):
-    # A field to hold the 'Id' and 'Changekey' attributes on 'ItemId' type items. There is no guaranteed max length,
-    # but we can assume 512 bytes in practice. See https://msdn.microsoft.com/en-us/library/office/dn605828(v=exchg.150)
+    """A field to hold the 'Id' and 'Changekey' attributes on 'ItemId' type items. There is no guaranteed max length,
+    but we can assume 512 bytes in practice. See
+    https://docs.microsoft.com/en-us/exchange/client-developer/exchange-web-services/ews-identifiers-in-exchange
+    """
     def __init__(self, *args, **kwargs):
         super(IdField, self).__init__(*args, **kwargs)
         self.max_length = 512  # This is above the normal 255 limit, but this is actually an attribute, not a field
@@ -661,31 +743,35 @@ class IdField(CharField):
 class CharListField(CharField):
     is_list = True
 
+    def list_elem_tag(self):
+        return '{%s}String' % self.namespace
+
     def from_xml(self, elem, account):
         iter_elem = elem.find(self.response_tag())
         if iter_elem is not None:
-            return get_xml_attrs(iter_elem, '{%s}String' % TNS)
+            return get_xml_attrs(iter_elem, self.list_elem_tag())
         return self.default
 
 
 class URIField(TextField):
-    # Helper to mark strings that must conform to xsd:anyURI
-    # If we want an URI validator, see http://stackoverflow.com/questions/14466585/is-this-regex-correct-for-xsdanyuri
+    """Helper to mark strings that must conform to xsd:anyURI
+    If we want an URI validator, see http://stackoverflow.com/questions/14466585/is-this-regex-correct-for-xsdanyuri
+    """
     pass
 
 
 class EmailAddressField(CharField):
-    # A helper class used for email address string that we can use for email validation
+    """A helper class used for email address string that we can use for email validation"""
     pass
 
 
 class CultureField(CharField):
-    # Helper to mark strings that are # RFC 1766 culture values.
+    """Helper to mark strings that are # RFC 1766 culture values."""
     pass
 
 
 class Choice(object):
-    """ Implements versioned choices for the ChoiceField field"""
+    """Implements versioned choices for the ChoiceField field"""
     def __init__(self, value, supported_from=None):
         self.value = value
         self.supported_from = supported_from
@@ -717,7 +803,7 @@ class ChoiceField(CharField):
             value, self.name, ', '.join(self.supported_choices(version=version))))
 
     def supported_choices(self, version=None):
-        return {c.value for c in self.choices if c.supports_version(version)}
+        return list(c.value for c in self.choices if c.supports_version(version))
 
 
 FREE_BUSY_CHOICES = [Choice('Free'), Choice('Tentative'), Choice('Busy'), Choice('OOF'), Choice('NoData'),
@@ -841,8 +927,6 @@ class OccurrenceListField(OccurrenceField):
 
 
 class MessageHeaderField(EWSElementListField):
-    is_complex = True
-
     def __init__(self, *args, **kwargs):
         from .properties import MessageHeader
         kwargs['value_cls'] = MessageHeader
@@ -850,7 +934,7 @@ class MessageHeaderField(EWSElementListField):
 
 
 class BaseEmailField(EWSElementField):
-    # A base class for EWSElement classes that have an 'email_address' field that we want to provide helpers for
+    """A base class for EWSElement classes that have an 'email_address' field that we want to provide helpers for"""
 
     is_complex = True  # FindItem only returns the name, not the email address
 
@@ -867,23 +951,31 @@ class BaseEmailField(EWSElementField):
         if sub_elem is not None:
             if self.field_uri is not None:
                 # We want the nested Mailbox, not the wrapper element
-                return self.value_cls.from_xml(elem=sub_elem.find(self.value_cls.response_tag()), account=account)
+                nested_elem = sub_elem.find(self.value_cls.response_tag())
+                if nested_elem is None:
+                    raise ValueError(
+                        'Expected XML element %r missing on field %r' % (self.value_cls.response_tag(), self.name)
+                    )
+                return self.value_cls.from_xml(elem=nested_elem, account=account)
             return self.value_cls.from_xml(elem=sub_elem, account=account)
         return self.default
 
 
 class EmailField(BaseEmailField):
-    is_complex = True  # FindItem only returns the name, not the email address
-
     def __init__(self, *args, **kwargs):
         from .properties import Email
         kwargs['value_cls'] = Email
         super(EmailField, self).__init__(*args, **kwargs)
 
 
-class MailboxField(BaseEmailField):
-    is_complex = True  # FindItem only returns the name, not the email address
+class RecipientAddressField(BaseEmailField):
+    def __init__(self, *args, **kwargs):
+        from .properties import RecipientAddress
+        kwargs['value_cls'] = RecipientAddress
+        super(RecipientAddressField, self).__init__(*args, **kwargs)
 
+
+class MailboxField(BaseEmailField):
     def __init__(self, *args, **kwargs):
         from .properties import Mailbox
         kwargs['value_cls'] = Mailbox
@@ -891,8 +983,6 @@ class MailboxField(BaseEmailField):
 
 
 class MailboxListField(EWSElementListField):
-    is_complex = True
-
     def __init__(self, *args, **kwargs):
         from .properties import Mailbox
         kwargs['value_cls'] = Mailbox
@@ -905,8 +995,6 @@ class MailboxListField(EWSElementListField):
 
 
 class MemberListField(EWSElementListField):
-    is_complex = True
-
     def __init__(self, *args, **kwargs):
         from .properties import Member
         kwargs['value_cls'] = Member
@@ -922,8 +1010,6 @@ class MemberListField(EWSElementListField):
 
 
 class AttendeesField(EWSElementListField):
-    is_complex = True
-
     def __init__(self, *args, **kwargs):
         from .properties import Attendee
         kwargs['value_cls'] = Attendee
@@ -938,8 +1024,6 @@ class AttendeesField(EWSElementListField):
 
 
 class AttachmentField(EWSElementListField):
-    is_complex = True
-
     def __init__(self, *args, **kwargs):
         from .attachments import Attachment
         kwargs['value_cls'] = Attachment
@@ -951,7 +1035,7 @@ class AttachmentField(EWSElementListField):
         # Look for both FileAttachment and ItemAttachment
         if iter_elem is not None:
             attachments = []
-            for att_type in (FileAttachment, ItemAttachment):
+            for att_type in (ItemAttachment, FileAttachment):
                 attachments.extend(
                     [att_type.from_xml(elem=e, account=account) for e in iter_elem.findall(att_type.response_tag())]
                 )
@@ -960,7 +1044,7 @@ class AttachmentField(EWSElementListField):
 
 
 class LabelField(ChoiceField):
-    # A field to hold the label on an IndexedElement
+    """A field to hold the label on an IndexedElement"""
     def __init__(self, *args, **kwargs):
         super(LabelField, self).__init__(*args, **kwargs)
         self.is_attribute = True
@@ -970,8 +1054,10 @@ class LabelField(ChoiceField):
 
 
 class SubField(Field):
+    namespace = TNS
+
     # A field to hold the value on an SingleFieldIndexedElement
-    value_cls = string_type
+    value_cls = string_types[0]
 
     def from_xml(self, elem, account):
         return elem.text
@@ -981,23 +1067,29 @@ class SubField(Field):
 
     @staticmethod
     def field_uri_xml(field_uri, label):
-        return create_element('t:IndexedFieldURI', FieldURI=field_uri, FieldIndex=label)
+        return create_element(
+            't:IndexedFieldURI',
+            attrs=OrderedDict([
+                ('FieldURI', field_uri),
+                ('FieldIndex', label),
+            ])
+        )
 
     def __hash__(self):
         return hash(self.name)
 
 
 class EmailSubField(SubField):
-    # A field to hold the value on an SingleFieldIndexedElement
-    value_cls = string_type
+    """A field to hold the value on an SingleFieldIndexedElement"""
+    value_cls = string_types[0]
 
     def from_xml(self, elem, account):
         return elem.text or elem.get('Name')  # Sometimes elem.text is empty. Exchange saves the same in 'Name' attr
 
 
 class NamedSubField(SubField):
-    # A field to hold the value on an MultiFieldIndexedElement
-    value_cls = string_type
+    """A field to hold the value on an MultiFieldIndexedElement"""
+    value_cls = string_types[0]
 
     def __init__(self, *args, **kwargs):
         self.field_uri = kwargs.pop('field_uri')
@@ -1017,34 +1109,24 @@ class NamedSubField(SubField):
         return set_xml_value(field_elem, value, version=version)
 
     def field_uri_xml(self, field_uri, label):
-        return create_element('t:IndexedFieldURI', FieldURI='%s:%s' % (field_uri, self.field_uri), FieldIndex=label)
+        return create_element(
+            't:IndexedFieldURI',
+            attrs=OrderedDict([
+                ('FieldURI', '%s:%s' % (field_uri, self.field_uri)),
+                ('FieldIndex', label),
+            ])
+        )
 
     def request_tag(self):
         return 't:%s' % self.field_uri
 
     def response_tag(self):
-        return '{%s}%s' % (TNS, self.field_uri)
+        return '{%s}%s' % (self.namespace, self.field_uri)
 
 
-class IndexedField(FieldURIField):
+class IndexedField(EWSElementField):
     PARENT_ELEMENT_NAME = None
 
-    def __init__(self, *args, **kwargs):
-        self.value_cls = kwargs.pop('value_cls')
-        super(IndexedField, self).__init__(*args, **kwargs)
-
-    def from_xml(self, elem, account):
-        if self.is_list:
-            iter_elem = elem.find(self.response_tag())
-            if iter_elem is not None:
-                return [self.value_cls.from_xml(elem=e, account=account)
-                        for e in iter_elem.findall(self.value_cls.response_tag())]
-        else:
-            sub_elem = elem.find(self.response_tag())
-            if sub_elem is not None:
-                return self.value_cls.from_xml(elem=sub_elem, account=account)
-        return self.default
-
     def to_xml(self, value, version):
         return set_xml_value(create_element('t:%s' % self.PARENT_ELEMENT_NAME), value, version)
 
@@ -1054,7 +1136,7 @@ class IndexedField(FieldURIField):
 
     @classmethod
     def response_tag(cls):
-        return '{%s}%s' % (TNS, cls.PARENT_ELEMENT_NAME)
+        return '{%s}%s' % (cls.namespace, cls.PARENT_ELEMENT_NAME)
 
     def __hash__(self):
         return hash(self.field_uri)
@@ -1089,7 +1171,7 @@ class PhoneNumberField(IndexedField):
 
 
 class PhysicalAddressField(IndexedField):
-    # MSDN: https://msdn.microsoft.com/en-us/library/office/aa564323(v=exchg.150).aspx
+    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/entry-physicaladdress"""
     is_list = True
 
     PARENT_ELEMENT_NAME = 'PhysicalAddresses'
@@ -1140,13 +1222,7 @@ class ExtendedPropertyField(Field):
     def from_xml(self, elem, account):
         extended_properties = elem.findall(self.value_cls.response_tag())
         for extended_property in extended_properties:
-            extended_field_uri = extended_property.find('{%s}ExtendedFieldURI' % TNS)
-            match = True
-            for k, v in self.value_cls.properties_map().items():
-                if extended_field_uri.get(k) != v:
-                    match = False
-                    break
-            if match:
+            if self.value_cls.is_property_instance(extended_property):
                 return self.value_cls.from_xml(elem=extended_property, account=account)
         return self.default
 
@@ -1184,11 +1260,22 @@ class ItemField(FieldURIField):
         return value.to_xml(version=version)
 
 
+class UnknownEntriesField(CharListField):
+    def list_elem_tag(self):
+        return '{%s}UnknownEntry' % self.namespace
+
+
+class PermissionSetField(EWSElementField):
+    is_complex = True
+
+    def __init__(self, *args, **kwargs):
+        from .properties import PermissionSet
+        kwargs['value_cls'] = PermissionSet
+        super(PermissionSetField, self).__init__(*args, **kwargs)
+
+
 class EffectiveRightsField(EWSElementField):
     def __init__(self, *args, **kwargs):
         from .properties import EffectiveRights
         kwargs['value_cls'] = EffectiveRights
         super(EffectiveRightsField, self).__init__(*args, **kwargs)
-
-    def from_xml(self, elem, account):
-        return super(EffectiveRightsField, self).from_xml(elem=elem, account=account)
