@@ -7,24 +7,19 @@ import logging
 
 from future.utils import python_2_unicode_compatible
 
-from .items import CalendarItem, IdOnly
+from .errors import MultipleObjectsReturned, DoesNotExist
+from .items import CalendarItem, ID_ONLY
 from .fields import FieldPath, FieldOrder
+from .properties import InvalidField
 from .restriction import Q
+from .services import CHUNK_SIZE
 from .version import EXCHANGE_2010
 
 log = logging.getLogger(__name__)
 
 
-class MultipleObjectsReturned(Exception):
-    pass
-
-
-class DoesNotExist(Exception):
-    pass
-
-
 class SearchableMixIn(object):
-    # Implements a search API for inheritance
+    """Implements a search API for inheritance"""
     def get(self, *args, **kwargs):
         raise NotImplementedError()
 
@@ -77,10 +72,11 @@ class QuerySet(SearchableMixIn):
         self.calendar_view = None
         self.page_size = None
         self.max_items = None
+        self.offset = 0
 
         self._cache = None
 
-    def copy(self):
+    def _copy_self(self):
         # When we copy a queryset where the cache has already been filled, we don't copy the cache. Thus, a copied
         # queryset will fetch results from the server again.
         #
@@ -107,39 +103,41 @@ class QuerySet(SearchableMixIn):
         new_qs.calendar_view = self.calendar_view
         new_qs.page_size = self.page_size
         new_qs.max_items = self.max_items
+        new_qs.offset = self.offset
         return new_qs
 
     @property
     def is_cached(self):
         return self._cache is not None
 
-    def _get_field_path(self, s):
+    def _get_field_path(self, field_path):
         from .items import Persona
         if self.request_type == self.PERSONA:
-            return FieldPath(field=Persona.get_field_by_fieldname(s))
+            return FieldPath(field=Persona.get_field_by_fieldname(field_path))
         for folder in self.folder_collection:
             try:
-                return FieldPath.from_string(s, folder=folder)
-            except ValueError:
+                return FieldPath.from_string(field_path=field_path, folder=folder)
+            except InvalidField:
                 pass
-        raise ValueError("Unknown fieldname '%s' on folders '%s'" % (s, self.folder_collection.folders))
+        raise InvalidField("Unknown field path %r on folders %s" % (field_path, self.folder_collection.folders))
 
-    def _get_field_order(self, s):
+    def _get_field_order(self, field_path):
         from .items import Persona
         if self.request_type == self.PERSONA:
-            field_path = FieldPath(field=Persona.get_field_by_fieldname(s.lstrip('-')))
-            reverse = s.startswith('-')
-            return FieldOrder(field_path=field_path, reverse=reverse)
+            return FieldOrder(
+                field_path=FieldPath(field=Persona.get_field_by_fieldname(field_path.lstrip('-'))),
+                reverse=field_path.startswith('-'),
+            )
         for folder in self.folder_collection:
             try:
-                return FieldOrder.from_string(s, folder=folder)
-            except ValueError:
+                return FieldOrder.from_string(field_path=field_path, folder=folder)
+            except InvalidField:
                 pass
-        raise ValueError("Unknown fieldname '%s' on folders '%s'" % (s, self.folder_collection.folders))
+        raise InvalidField("Unknown field path %r on folders %s" % (field_path, self.folder_collection.folders))
 
     @property
     def _item_id_field(self):
-        return self._get_field_path('item_id')
+        return self._get_field_path('id')
 
     @property
     def _changekey_field(self):
@@ -167,6 +165,14 @@ class QuerySet(SearchableMixIn):
                 additional_fields.add(FieldPath(field=end_tz_field))
         return additional_fields
 
+    def _format_items(self, items, return_format):
+        return {
+            self.VALUES: self._as_values,
+            self.VALUES_LIST: self._as_values_list,
+            self.FLAT: self._as_flat_values_list,
+            self.NONE: self._as_items,
+        }[return_format](items)
+
     def _query(self):
         from .folders import SHALLOW
         from .items import Persona
@@ -178,12 +184,11 @@ class QuerySet(SearchableMixIn):
                 ) if not f.is_complex}
                 complex_fields_requested = False
             else:
-                additional_fields = {FieldPath(field=f) for f in self.folder_collection.allowed_fields()}
+                additional_fields = {FieldPath(field=f) for f in self.folder_collection.allowed_item_fields()}
                 complex_fields_requested = True
         else:
             additional_fields = self._additional_fields()
-            complex_fields_requested = \
-                bool(set(f.field for f in additional_fields) & self.folder_collection.complex_fields())
+            complex_fields_requested = any(f.field.is_complex for f in additional_fields)
 
         # EWS can do server-side sorting on multiple fields. A caveat is that server-side sorting is not supported
         # for calendar views. In this case, we do all the sorting client-side.
@@ -205,29 +210,30 @@ class QuerySet(SearchableMixIn):
         if self.request_type == self.PERSONA:
             if len(self.folder_collection) != 1:
                 raise ValueError('Personas can only be queried on a single folder')
-            folder = list(self.folder_collection)[0]
-            items = folder.find_people(
+            items = list(self.folder_collection)[0].find_people(
                 self.q,
-                shape=IdOnly,
+                shape=ID_ONLY,
                 depth=SHALLOW,
                 additional_fields=additional_fields,
                 order_fields=order_fields,
                 page_size=self.page_size,
                 max_items=self.max_items,
+                offset=self.offset,
             )
         else:
             find_item_kwargs = dict(
-                shape=IdOnly,  # Always use IdOnly here, because AllProperties doesn't actually get *all* properties
+                shape=ID_ONLY,  # Always use IdOnly here, because AllProperties doesn't actually get *all* properties
                 additional_fields=additional_fields,
                 order_fields=order_fields,
                 calendar_view=self.calendar_view,
                 page_size=self.page_size,
                 max_items=self.max_items,
+                offset=self.offset,
             )
 
             if complex_fields_requested:
                 # The FindItem service does not support complex field types. Tell find_items() to return
-                # (item_id, changekey) tuples, and pass that to fetch().
+                # (id, changekey) tuples, and pass that to fetch().
                 find_item_kwargs['additional_fields'] = None
                 items = self.folder_collection.account.fetch(
                     ids=self.folder_collection.find_items(self.q, **find_item_kwargs),
@@ -236,9 +242,9 @@ class QuerySet(SearchableMixIn):
                 )
             else:
                 if not additional_fields:
-                    # If additional_fields is the empty set, we only requested item_id and changekey fields. We can then
-                    # take a shortcut by using (shape=IdOnly, additional_fields=None) to tell find_items() to return
-                    # (item_id, changekey) tuples. We'll post-process those later.
+                    # If additional_fields is the empty set, we only requested ID and changekey fields. We can then
+                    # take a shortcut by using (shape=ID_ONLY, additional_fields=None) to tell find_items() to return
+                    # (id, changekey) tuples. We'll post-process those later.
                     find_item_kwargs['additional_fields'] = None
                 items = self.folder_collection.find_items(self.q, **find_item_kwargs)
 
@@ -248,22 +254,9 @@ class QuerySet(SearchableMixIn):
         # Resort to client-side sorting of the order_by fields. This is greedy. Sorting in Python is stable, so when
         # sorting on multiple fields, we can just do a sort on each of the requested fields in reverse order. Reverse
         # each sort operation if the field was marked as such.
-        def get_value_or_default(item, field_order):
-            # Python can only sort values when <, > and = are implemented for the two types. Try as best we can to sort
-            # items, even when the item may have a None value for the field in question, or when the item is an
-            # Exception. If the field to be sorted by does not have a default value, there's really nothing we can do
-            # about it; we'll eventually raise a TypeError. If it does, we sort all None values and exceptions as the
-            # default value.
-            if isinstance(item, Exception):
-                return field_order.field_path.field.default
-            val = field_order.field_path.get_value(item)
-            if val is None:
-                return field_order.field_path.field.default
-            return val
-
         for f in reversed(self.order_fields):
             try:
-                items = sorted(items, key=lambda i: get_value_or_default(i, f), reverse=f.reverse)
+                items = sorted(items, key=lambda i: _get_value_or_default(i, f), reverse=f.reverse)
             except TypeError as e:
                 if 'unorderable types' not in e.args[0]:
                     raise
@@ -274,14 +267,8 @@ class QuerySet(SearchableMixIn):
         if not extra_order_fields:
             return items
 
-        # Nullify the fields we only needed for sorting. Make sure to handle exceptions.
-        def clean_item(i):
-            if isinstance(i, Exception):
-                return i
-            for f in extra_order_fields:
-                setattr(i, f.field.name, None)
-            return i
-        return (clean_item(i) for i in items)
+        # Nullify the fields we only needed for sorting before returning
+        return (_rinse_item(i, extra_order_fields) for i in items)
 
     def __iter__(self):
         # Fill cache if this is the first iteration. Return an iterator over the results. Make this non-greedy by
@@ -300,13 +287,7 @@ class QuerySet(SearchableMixIn):
 
         log.debug('Initializing cache')
         _cache = []
-        result_formatter = {
-            self.VALUES: self._as_values,
-            self.VALUES_LIST: self._as_values_list,
-            self.FLAT: self._as_flat_values_list,
-            self.NONE: self._as_items,
-        }[self.return_format]
-        for val in result_formatter(self._query()):
+        for val in self._format_items(items=self._query(), return_format=self.return_format):
             _cache.append(val)
             yield val
         self._cache = _cache
@@ -320,8 +301,6 @@ class QuerySet(SearchableMixIn):
     def __getitem__(self, idx_or_slice):
         # Support indexing and slicing. This is non-greedy when possible (slicing start, stop and step are not negative,
         # and we're ordering on at most one field), and will only fill the cache if the entire query is iterated.
-        # TODO: We could optimize this for large indexes or slices (e.g. [999] or [999:1002]) by letting the FindItem
-        # service expose the 'offset' value, so we don't need to get the first 999 items.
         if isinstance(idx_or_slice, int):
             return self._getitem_idx(idx_or_slice)
         return self._getitem_slice(idx_or_slice)
@@ -333,16 +312,15 @@ class QuerySet(SearchableMixIn):
             # Support negative indexes by reversing the queryset and negating the index value
             reverse_idx = -(idx+1)
             return self.reverse()[reverse_idx]
-        else:
-            if not self.is_cached and idx < 100:
-                # If idx is small, optimize a bit by setting self.page_size to only get as many items as strictly needed
-                self.page_size = idx + 1
-                self.max_items = idx + 1
-            # Support non-negative indexes by consuming the iterator up to the index
-            for i, val in enumerate(self.__iter__()):
-                if i == idx:
-                    return val
-            raise IndexError()
+        # Optimize by setting an exact offset and fetching only 1 item
+        new_qs = self._copy_self()
+        new_qs.max_items = 1
+        new_qs.page_size = 1
+        new_qs.offset = idx
+        # The iterator will return at most 1 item
+        for item in new_qs.__iter__():
+            return item
+        raise IndexError()
 
     def _getitem_slice(self, s):
         if ((s.start or 0) < 0) or ((s.stop or 0) < 0) or ((s.step or 0) < 0):
@@ -350,139 +328,91 @@ class QuerySet(SearchableMixIn):
             # query result, and then slice on the cache.
             list(self.__iter__())
             return self._cache[s]
-        if not self.is_cached and s.stop is not None and s.stop < 100:
-            # If the range is small, optimize a bit by setting self.page_size to only get as many items as strictly
-            # needed.
-            self.page_size = s.stop
-            self.max_items = s.stop
-        return islice(self.__iter__(), s.start, s.stop, s.step)
+        if self.is_cached:
+            return islice(self.__iter__(), s.start, s.stop, s.step)
+        # Optimize by setting an exact offset and max_items value
+        new_qs = self._copy_self()
+        if s.start is not None and s.stop is not None:
+            new_qs.offset = s.start
+            new_qs.max_items = s.stop - s.start
+        elif s.start is not None:
+            new_qs.offset = s.start
+        elif s.stop is not None:
+            new_qs.max_items = s.stop
+        if new_qs.page_size is None and new_qs.max_items is not None and new_qs.max_items < CHUNK_SIZE:
+            new_qs.page_size = new_qs.max_items
+        return islice(new_qs.__iter__(), None, None, s.step)
+
+    def _item_yielder(self, iterable, item_func, id_only_func, changekey_only_func, id_and_changekey_func):
+        # Transforms results from the server according to the given transform functions. Makes sure to pass on
+        # Exception instances unaltered.
+        if self.only_fields:
+            has_non_attribute_fields = bool({f for f in self.only_fields if not f.field.is_attribute})
+        else:
+            has_non_attribute_fields = True
+        if not has_non_attribute_fields:
+            # _query() will return an iterator of (id, changekey) tuples
+            if self._changekey_field not in self.only_fields:
+                transform_func = id_only_func
+            elif self._item_id_field not in self.only_fields:
+                transform_func = changekey_only_func
+            else:
+                transform_func = id_and_changekey_func
+            for i in iterable:
+                if isinstance(i, Exception):
+                    yield i
+                    continue
+                yield transform_func(*i)
+            return
+        for i in iterable:
+            if isinstance(i, Exception):
+                yield i
+                continue
+            yield item_func(i)
 
     def _as_items(self, iterable):
         from .items import Item
-        if self.only_fields:
-            has_non_attribute_fields = bool({f for f in self.only_fields if not f.field.is_attribute})
-            if not has_non_attribute_fields:
-                # _query() will return an iterator of (item_id, changekey) tuples
-                if self._changekey_field not in self.only_fields:
-                    for i in iterable:
-                        if isinstance(i, Exception):
-                            yield i
-                            continue
-                        item_id, changekey = i
-                        yield Item(item_id=item_id)
-                elif self._item_id_field not in self.only_fields:
-                    for i in iterable:
-                        if isinstance(i, Exception):
-                            yield i
-                            continue
-                        item_id, changekey = i
-                        yield Item(changekey=changekey)
-                else:
-                    for i in iterable:
-                        if isinstance(i, Exception):
-                            yield i
-                            continue
-                        item_id, changekey = i
-                        yield Item(item_id=item_id, changekey=changekey)
-                return
-        for i in iterable:
-            yield i
+        return self._item_yielder(
+            iterable=iterable,
+            item_func=lambda i: i,
+            id_only_func=lambda item_id, changekey: Item(id=item_id),
+            changekey_only_func=lambda item_id, changekey: Item(changekey=changekey),
+            id_and_changekey_func=lambda item_id, changekey: Item(id=item_id, changekey=changekey),
+        )
 
     def _as_values(self, iterable):
         if not self.only_fields:
             raise ValueError('values() requires at least one field name')
-        has_non_attribute_fields = bool({f for f in self.only_fields if not f.field.is_attribute})
-        if not has_non_attribute_fields:
-            # _query() will return an iterator of (item_id, changekey) tuples
-            if self._changekey_field not in self.only_fields:
-                for i in iterable:
-                    if isinstance(i, Exception):
-                        yield i
-                        continue
-                    item_id, changekey = i
-                    yield {'item_id': item_id}
-            elif self._item_id_field not in self.only_fields:
-                for i in iterable:
-                    if isinstance(i, Exception):
-                        yield i
-                        continue
-                    item_id, changekey = i
-                    yield {'changekey': changekey}
-            else:
-                for i in iterable:
-                    if isinstance(i, Exception):
-                        yield i
-                        continue
-                    item_id, changekey = i
-                    yield {'item_id': item_id, 'changekey': changekey}
-            return
-        for i in iterable:
-            if isinstance(i, Exception):
-                yield i
-                continue
-            yield {f.path: f.get_value(i) for f in self.only_fields}
+        return self._item_yielder(
+            iterable=iterable,
+            item_func=lambda i: {f.path: f.get_value(i) for f in self.only_fields},
+            id_only_func=lambda item_id, changekey: {'id': item_id},
+            changekey_only_func=lambda item_id, changekey: {'changekey': changekey},
+            id_and_changekey_func=lambda item_id, changekey: {'id': item_id, 'changekey': changekey},
+        )
 
     def _as_values_list(self, iterable):
         if not self.only_fields:
             raise ValueError('values_list() requires at least one field name')
-        has_non_attribute_fields = bool({f for f in self.only_fields if not f.field.is_attribute})
-        if not has_non_attribute_fields:
-            # _query() will return an iterator of (item_id, changekey) tuples
-            if self._changekey_field not in self.only_fields:
-                for i in iterable:
-                    if isinstance(i, Exception):
-                        yield i
-                        continue
-                    item_id, changekey = i
-                    yield (item_id,)
-            elif self._item_id_field not in self.only_fields:
-                for i in iterable:
-                    if isinstance(i, Exception):
-                        yield i
-                        continue
-                    item_id, changekey = i
-                    yield (changekey,)
-            else:
-                for i in iterable:
-                    if isinstance(i, Exception):
-                        yield i
-                        continue
-                    item_id, changekey = i
-                    yield (item_id, changekey)
-            return
-        for i in iterable:
-            if isinstance(i, Exception):
-                yield i
-                continue
-            yield tuple(f.get_value(i) for f in self.only_fields)
+        return self._item_yielder(
+            iterable=iterable,
+            item_func=lambda i: tuple(f.get_value(i) for f in self.only_fields),
+            id_only_func=lambda item_id, changekey: (item_id,),
+            changekey_only_func=lambda item_id, changekey: (changekey,),
+            id_and_changekey_func=lambda item_id, changekey: (item_id, changekey),
+        )
 
     def _as_flat_values_list(self, iterable):
         if not self.only_fields or len(self.only_fields) != 1:
             raise ValueError('flat=True requires exactly one field name')
         flat_field_path = self.only_fields[0]
-        if flat_field_path == self._item_id_field:
-            # _query() will return an iterator of (item_id, changekey) tuples
-            for i in iterable:
-                if isinstance(i, Exception):
-                    yield i
-                    continue
-                item_id, changekey = i
-                yield item_id
-            return
-        if flat_field_path == self._changekey_field:
-            # _query() will return an iterator of (item_id, changekey) tuples
-            for i in iterable:
-                if isinstance(i, Exception):
-                    yield i
-                    continue
-                item_id, changekey = i
-                yield changekey
-            return
-        for i in iterable:
-            if isinstance(i, Exception):
-                yield i
-                continue
-            yield flat_field_path.get_value(i)
+        return self._item_yielder(
+            iterable=iterable,
+            item_func=flat_field_path.get_value,
+            id_only_func=lambda item_id, changekey: item_id,
+            changekey_only_func=lambda item_id, changekey: changekey,
+            id_and_changekey_func=None,  # Can never be called
+        )
 
     ###############################
     #
@@ -497,32 +427,32 @@ class QuerySet(SearchableMixIn):
     #
     def all(self):
         """ Return everything, without restrictions """
-        new_qs = self.copy()
+        new_qs = self._copy_self()
         return new_qs
 
     def none(self):
         """ Return a query that is guaranteed to be empty  """
-        new_qs = self.copy()
+        new_qs = self._copy_self()
         new_qs.q = None
         return new_qs
 
     def filter(self, *args, **kwargs):
         """ Return everything that matches these search criteria """
-        new_qs = self.copy()
+        new_qs = self._copy_self()
         q = Q(*args, **kwargs)
         new_qs.q = q if new_qs.q is None else new_qs.q & q
         return new_qs
 
     def exclude(self, *args, **kwargs):
         """ Return everything that does NOT match these search criteria """
-        new_qs = self.copy()
+        new_qs = self._copy_self()
         q = ~Q(*args, **kwargs)
         new_qs.q = q if new_qs.q is None else new_qs.q & q
         return new_qs
 
     def people(self):
         """ Changes the queryset to search the folder for Personas instead of Items """
-        new_qs = self.copy()
+        new_qs = self._copy_self()
         new_qs.request_type = self.PERSONA
         return new_qs
 
@@ -532,7 +462,7 @@ class QuerySet(SearchableMixIn):
             only_fields = tuple(self._get_field_path(arg) for arg in args)
         except ValueError as e:
             raise ValueError("%s in only()" % e.args[0])
-        new_qs = self.copy()
+        new_qs = self._copy_self()
         new_qs.only_fields = only_fields
         return new_qs
 
@@ -544,7 +474,7 @@ class QuerySet(SearchableMixIn):
             order_fields = tuple(self._get_field_order(arg) for arg in args)
         except ValueError as e:
             raise ValueError("%s in order_by()" % e.args[0])
-        new_qs = self.copy()
+        new_qs = self._copy_self()
         new_qs.order_fields = order_fields
         return new_qs
 
@@ -552,7 +482,7 @@ class QuerySet(SearchableMixIn):
         """ Return the entire query result in reverse order """
         if not self.order_fields:
             raise ValueError('Reversing only makes sense if there are order_by fields')
-        new_qs = self.copy()
+        new_qs = self._copy_self()
         for f in new_qs.order_fields:
             f.reverse = not f.reverse
         return new_qs
@@ -563,7 +493,7 @@ class QuerySet(SearchableMixIn):
             only_fields = tuple(self._get_field_path(arg) for arg in args)
         except ValueError as e:
             raise ValueError("%s in values()" % e.args[0])
-        new_qs = self.copy()
+        new_qs = self._copy_self()
         new_qs.only_fields = only_fields
         new_qs.return_format = self.VALUES
         return new_qs
@@ -582,7 +512,7 @@ class QuerySet(SearchableMixIn):
             only_fields = tuple(self._get_field_path(arg) for arg in args)
         except ValueError as e:
             raise ValueError("%s in values_list()" % e.args[0])
-        new_qs = self.copy()
+        new_qs = self._copy_self()
         new_qs.only_fields = only_fields
         new_qs.return_format = self.FLAT if flat else self.VALUES_LIST
         return new_qs
@@ -593,25 +523,24 @@ class QuerySet(SearchableMixIn):
     #
     ###########################
     def iterator(self):
-        """ Return the query result as an iterator, without caching the result. 'page_size' is the number of items to
-        fetch from the server per request. """
+        """ Return the query result as an iterator, without caching the result """
         if self.q is None:
             return []
         if self.is_cached:
             return self._cache
         # Return an iterator that doesn't bother with caching
-        return self._query()
+        return self._format_items(items=self._query(), return_format=self.return_format)
 
     def get(self, *args, **kwargs):
         """ Assume the query will return exactly one item. Return that item """
         if self.is_cached and not args and not kwargs:
             # We can only safely use the cache if get() is called without args
             items = self._cache
-        elif not args and set(kwargs.keys()) in ({'item_id'}, {'item_id', 'changekey'}):
-            # We allow calling get(item_id=..., changekey=...) to get a single item, but only if exactly these two
+        elif not args and set(kwargs.keys()) in ({'id'}, {'id', 'changekey'}):
+            # We allow calling get(id=..., changekey=...) to get a single item, but only if exactly these two
             # kwargs are present.
             account = self.folder_collection.account
-            item_id = self._item_id_field.field.clean(kwargs['item_id'], version=account.version)
+            item_id = self._item_id_field.field.clean(kwargs['id'], version=account.version)
             changekey = self._changekey_field.field.clean(kwargs.get('changekey'), version=account.version)
             items = list(account.fetch(ids=[(item_id, changekey)], only_fields=self.only_fields))
         else:
@@ -628,7 +557,7 @@ class QuerySet(SearchableMixIn):
         fetch from the server per request. We're only fetching the IDs, so keep it high"""
         if self.is_cached:
             return len(self._cache)
-        new_qs = self.copy()
+        new_qs = self._copy_self()
         new_qs.only_fields = tuple()
         new_qs.order_fields = None
         new_qs.return_format = self.NONE
@@ -639,35 +568,123 @@ class QuerySet(SearchableMixIn):
         """ Find out if the query contains any hits, with as little effort as possible """
         if self.is_cached:
             return len(self._cache) > 0
-        new_qs = self.copy()
+        new_qs = self._copy_self()
         new_qs.max_items = 1
         return new_qs.count(page_size=1) > 0
 
-    def delete(self, page_size=1000):
-        """ Delete the items matching the query, with as little effort as possible. 'page_size' is the number of items
-        to fetch and delete from the server per request. We're only fetching the IDs, so keep it high"""
-        from .items import ALL_OCCURRENCIES
-        if self.is_cached:
-            res = self.folder_collection.account.bulk_delete(
-                ids=self._cache,
-                affected_task_occurrences=ALL_OCCURRENCIES,
-                chunk_size=page_size,
-            )
-            self._cache = None  # Invalidate the cache after delete, regardless of the results
-            return res
-        new_qs = self.copy()
+    def _id_only_copy_self(self):
+        new_qs = self._copy_self()
         new_qs.only_fields = tuple()
         new_qs.order_fields = None
         new_qs.return_format = self.NONE
-        new_qs.page_size = page_size
-        return self.folder_collection.account.bulk_delete(
-            ids=new_qs,
-            affected_task_occurrences=ALL_OCCURRENCIES,
+        return new_qs
+
+    def delete(self, page_size=1000, **delete_kwargs):
+        """ Delete the items matching the query, with as little effort as possible. 'page_size' is the number of items
+        to fetch and delete per request. We're only fetching the IDs, so keep it high"""
+        if self.is_cached:
+            ids = self._cache
+        else:
+            ids = self._id_only_copy_self()
+            ids.page_size = page_size
+        res = self.folder_collection.account.bulk_delete(
+            ids=ids,
+            chunk_size=page_size,
+            **delete_kwargs
+        )
+        self._cache = None  # Invalidate the cache, regardless of the results
+        return res
+
+    def send(self, page_size=1000, **send_kwargs):
+        """ Send the items matching the query, with as little effort as possible. 'page_size' is the number of items
+        to fetch and send per request. We're only fetching the IDs, so keep it high"""
+        if self.is_cached:
+            ids = self._cache
+        else:
+            ids = self._id_only_copy_self()
+            ids.page_size = page_size
+        res = self.folder_collection.account.bulk_send(
+            ids=ids,
+            chunk_size=page_size,
+            **send_kwargs
+        )
+        self._cache = None  # Invalidate the cache, regardless of the results
+        return res
+
+    def copy(self, to_folder, page_size=1000, **copy_kwargs):
+        """ Copy the items matching the query, with as little effort as possible. 'page_size' is the number of items
+        to fetch and copy per request. We're only fetching the IDs, so keep it high"""
+        if self.is_cached:
+            ids = self._cache
+        else:
+            ids = self._id_only_copy_self()
+            ids.page_size = page_size
+        res = self.folder_collection.account.bulk_copy(
+            ids=ids,
+            to_folder=to_folder,
+            chunk_size=page_size,
+            **copy_kwargs
+        )
+        self._cache = None  # Invalidate the cache, regardless of the results
+        return res
+
+    def move(self, to_folder, page_size=1000):
+        """ Move the items matching the query, with as little effort as possible. 'page_size' is the number of items
+        to fetch and move per request. We're only fetching the IDs, so keep it high"""
+        if self.is_cached:
+            ids = self._cache
+        else:
+            ids = self._id_only_copy_self()
+            ids.page_size = page_size
+        res = self.folder_collection.account.bulk_move(
+            ids=ids,
+            to_folder=to_folder,
             chunk_size=page_size,
         )
+        self._cache = None  # Invalidate the cache after delete, regardless of the results
+        return res
+
+    def archive(self, to_folder, page_size=1000):
+        """ Archive the items matching the query, with as little effort as possible. 'page_size' is the number of items
+        to fetch and move per request. We're only fetching the IDs, so keep it high"""
+        if self.is_cached:
+            ids = self._cache
+        else:
+            ids = self._id_only_copy_self()
+            ids.page_size = page_size
+        res = self.folder_collection.account.bulk_archive(
+            ids=ids,
+            to_folder=to_folder,
+            chunk_size=page_size,
+        )
+        self._cache = None  # Invalidate the cache after delete, regardless of the results
+        return res
 
     def __str__(self):
         fmt_args = [('q', str(self.q)), ('folders', '[%s]' % ', '.join(str(f) for f in self.folder_collection.folders))]
         if self.is_cached:
             fmt_args.append(('len', str(len(self))))
         return self.__class__.__name__ + '(%s)' % ', '.join('%s=%s' % (k, v) for k, v in fmt_args)
+
+
+def _get_value_or_default(item, field_order):
+    # Python can only sort values when <, > and = are implemented for the two types. Try as best we can to sort
+    # items, even when the item may have a None value for the field in question, or when the item is an
+    # Exception. If the field to be sorted by does not have a default value, there's really nothing we can do
+    # about it; we'll eventually raise a TypeError. If it does, we sort all None values and exceptions as the
+    # default value.
+    if isinstance(item, Exception):
+        return field_order.field_path.field.default
+    val = field_order.field_path.get_value(item)
+    if val is None:
+        return field_order.field_path.field.default
+    return val
+
+
+def _rinse_item(i, fields_to_nullify):
+    # Set fields in fields_to_nullify to None. Make sure to accept exceptions.
+    if isinstance(i, Exception):
+        return i
+    for f in fields_to_nullify:
+        setattr(i, f.field.name, None)
+    return i
