@@ -1,9 +1,12 @@
 # coding=utf-8
+import base64
+from collections import OrderedDict
 import logging
 
 from future.utils import python_2_unicode_compatible
 from six import string_types
 
+from .properties import InvalidField
 from .util import create_element, xml_to_str, value_to_xml_text, is_iterable
 from .version import EXCHANGE_2010
 
@@ -12,6 +15,8 @@ log = logging.getLogger(__name__)
 
 @python_2_unicode_compatible
 class Q(object):
+    """A class with an API similar to Django Q objects. Used to implemnt advanced filtering logic."""
+
     # Connection types
     AND = 'AND'
     OR = 'OR'
@@ -82,7 +87,7 @@ class Q(object):
         # Parse args which must be Q objects
         for q in args:
             if not isinstance(q, self.__class__):
-                raise ValueError("Non-keyword arg %r must be a Q object" % q)
+                raise ValueError("Non-keyword arg %r must be a Q instance" % q)
             if q.query_string:
                 raise ValueError(
                     'A query string cannot be combined with other restrictions (args: %r, kwargs: %r)' % (args, kwargs)
@@ -184,7 +189,7 @@ class Q(object):
         # Do some basic checks on the attributes, using a generic folder and no Exchange version restrictions. to_xml()
         # does a really good job of validating. There's no reason to replicate much of that here.
         from .folders import Folder
-        self.to_xml(folders=[Folder()], version=None)
+        self.to_xml(folders=[Folder()], version=None, applies_to=Restriction.ITEMS)
 
     @classmethod
     def _lookup_to_op(cls, lookup):
@@ -230,7 +235,7 @@ class Q(object):
             raise ValueError("'op' %s must be one of %s" % (op, valid_ops))
 
         # For description of Contains attribute values, see
-        #     https://msdn.microsoft.com/en-us/library/office/aa580702(v=exchg.150).aspx
+        #     https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/contains
         #
         # Possible ContainmentMode values:
         #     FullString, Prefixed, Substring, PrefixOnWords, ExactPhrase
@@ -262,7 +267,13 @@ class Q(object):
             compare_mode = 'IgnoreCase'
         else:
             compare_mode = 'Exact'
-        return create_element('t:Contains', ContainmentMode=match_mode, ContainmentComparison=compare_mode)
+        return create_element(
+            't:Contains',
+            attrs=OrderedDict([
+                ('ContainmentMode', match_mode),
+                ('ContainmentComparison', compare_mode),
+            ])
+        )
 
     def is_leaf(self):
         return not self.children
@@ -276,7 +287,7 @@ class Q(object):
         if self.query_string:
             return self.query_string
         if self.is_leaf():
-            expr = '%s %s %s' % (self.field_path, self.op, repr(self.value))
+            expr = '%s %s %r' % (self.field_path, self.op, self.value)
         else:
             # Sort children by field name so we get stable output (for easier testing). Children should never be empty.
             expr = (' %s ' % (self.AND if self.conn_type == self.NOT else self.conn_type)).join(
@@ -290,7 +301,7 @@ class Q(object):
             return self.conn_type + ' (%s)' % expr
         return expr
 
-    def to_xml(self, folders, version):
+    def to_xml(self, folders, version, applies_to):
         if self.query_string:
             if version.build < EXCHANGE_2010:
                 raise NotImplementedError('QueryString filtering is only supported for Exchange 2010 servers and later')
@@ -298,7 +309,7 @@ class Q(object):
             elem.text = self.query_string
             return elem
         # Translate this Q object to a valid Restriction XML tree
-        elem = self.xml_elem(folders=folders, version=version)
+        elem = self.xml_elem(folders=folders, version=version, applies_to=applies_to)
         if elem is None:
             return None
         restriction = create_element('m:Restriction')
@@ -330,11 +341,13 @@ class Q(object):
                 'Value %r for filter on field path "%s" must be a single value' % (self.value, self.field_path)
             )
 
-    def _validate_field_path(self, field_path, folder):
+    def _validate_field_path(self, field_path, folder, applies_to, version):
         from .indexed_properties import MultiFieldIndexedElement
-        if field_path.field not in folder.allowed_fields():
-            raise ValueError(
-                "'%s' is not a valid field when filtering on %s" % (field_path.field.name, folder.__class__.__name__))
+        if applies_to == Restriction.FOLDERS:
+            # This is a restriction on Folder fields
+            folder.validate_field(field=field_path.field, version=version)
+        else:
+            folder.validate_item_field(field=field_path.field)
         if not field_path.field.is_searchable:
             raise ValueError("EWS does not support filtering on field '%s'" % field_path.field.name)
         if field_path.subfield and not field_path.subfield.is_searchable:
@@ -342,18 +355,23 @@ class Q(object):
         if issubclass(field_path.field.value_cls, MultiFieldIndexedElement) and not field_path.subfield:
             raise ValueError("Field path '%s' must contain a subfield" % self.field_path)
 
-    def _get_field_path(self, folders):
+    def _get_field_path(self, folders, applies_to, version):
         # Convert the string field path to a real FieldPath object. The path is validated using the given folders.
         from .fields import FieldPath
         for folder in folders:
             try:
-                field_path = FieldPath.from_string(self.field_path, folder=folder)
-                self._validate_field_path(field_path=field_path, folder=folder)
-                break
+                if applies_to == Restriction.FOLDERS:
+                    # This is a restriction on Folder fields
+                    field = folder.get_field_by_fieldname(fieldname=self.field_path)
+                    field_path = FieldPath(field=field)
+                else:
+                    field_path = FieldPath.from_string(field_path=self.field_path, folder=folder)
             except ValueError:
                 continue
+            self._validate_field_path(field_path=field_path, folder=folder, applies_to=applies_to, version=version)
+            break
         else:
-            raise ValueError("Unknown fieldname '%s' on folders '%s'" % (self.field_path, folders))
+            raise InvalidField("Unknown field path %r on folders %s" % (self.field_path, folders))
         return field_path
 
     def _get_clean_value(self, field_path, version):
@@ -366,10 +384,11 @@ class Q(object):
         else:
             return clean_field.clean(value=self.value, version=version)
 
-    def xml_elem(self, folders, version):
+    def xml_elem(self, folders, version, applies_to):
         # Recursively build an XML tree structure of this Q object. If this is an empty leaf (the equivalent of Q()),
         # return None.
         from .indexed_properties import SingleFieldIndexedElement
+        from .extended_properties import ExtendedProperty
         # Don't check self.value just yet. We want to return error messages on the field path first, and then the value.
         # This is done in _get_field_path() and _get_clean_value(), respectively.
         self._check_integrity()
@@ -377,9 +396,12 @@ class Q(object):
             return None
         if self.is_leaf():
             elem = self._op_to_xml(self.op)
-            field_path = self._get_field_path(folders)
+            field_path = self._get_field_path(folders, applies_to=applies_to, version=version)
             clean_value = self._get_clean_value(field_path=field_path, version=version)
-            if issubclass(field_path.field.value_cls, SingleFieldIndexedElement) and not field_path.label:
+            if issubclass(field_path.field.value_cls, ExtendedProperty) and field_path.field.value_cls.is_binary_type():
+                # We need to base64-encode binary data
+                clean_value = base64.b64encode(clean_value.value).decode('ascii')
+            elif issubclass(field_path.field.value_cls, SingleFieldIndexedElement) and not field_path.label:
                 # We allow a filter shortcut of e.g. email_addresses__contains=EmailAddress(label='Foo', ...) instead of
                 # email_addresses__Foo_email_address=.... Set FieldPath label now so we can generate the field_uri.
                 field_path.label = clean_value.label
@@ -396,13 +418,13 @@ class Q(object):
                     elem.append(uriorconst)
         elif len(self.children) == 1:
             # We have only one child
-            elem = self.children[0].xml_elem(folders=folders, version=version)
+            elem = self.children[0].xml_elem(folders=folders, version=version, applies_to=applies_to)
         else:
             # We have multiple children. If conn_type is NOT, then group children with AND. We'll add the NOT later
             elem = self._conn_to_xml(self.AND if self.conn_type == self.NOT else self.conn_type)
             # Sort children by field name so we get stable output (for easier testing). Children should never be empty
             for c in sorted(self.children, key=lambda i: i.field_path or ''):
-                elem.append(c.xml_elem(folders=folders, version=version))
+                elem.append(c.xml_elem(folders=folders, version=version, applies_to=applies_to))
         if elem is None:
             return None  # Should not be necessary, but play safe
         if self.conn_type == self.NOT:
@@ -461,8 +483,8 @@ class Q(object):
     def __repr__(self):
         if self.is_leaf():
             if self.query_string:
-                return self.__class__.__name__ + '(%s)' % repr(self.query_string)
-            return self.__class__.__name__ + '(%s %s %s)' % (self.field_path, self.op, repr(self.value))
+                return self.__class__.__name__ + '(%r)' % self.query_string
+            return self.__class__.__name__ + '(%s %s %r)' % (self.field_path, self.op, self.value)
         sorted_children = tuple(sorted(self.children, key=lambda i: i.field_path or ''))
         if self.conn_type == self.NOT or len(self.children) > 1:
             return self.__class__.__name__ + repr((self.conn_type,) + sorted_children)
@@ -476,20 +498,28 @@ class Restriction(object):
 
     """
 
-    def __init__(self, q, folders):
+    # The type of item the restriction applies to
+    FOLDERS = 'folders'
+    ITEMS = 'items'
+    RESTRICTION_TYPES = (FOLDERS, ITEMS)
+
+    def __init__(self, q, folders, applies_to):
         if not isinstance(q, Q):
             raise ValueError("'q' value %r must be a Q instance" % q)
         if q.is_empty():
             raise ValueError("Q object must not be empty")
-        from .folders import Folder
+        from .folders import BaseFolder
         for folder in folders:
-            if not isinstance(folder, Folder):
+            if not isinstance(folder, BaseFolder):
                 raise ValueError("'folder' value %r must be a Folder instance" % folder)
+        if applies_to not in self.RESTRICTION_TYPES:
+            raise ValueError("'applies_to' must be one of %s" % (self.RESTRICTION_TYPES,))
         self.q = q
         self.folders = folders
+        self.applies_to = applies_to
 
     def to_xml(self, version):
-        return self.q.to_xml(folders=self.folders, version=version)
+        return self.q.to_xml(folders=self.folders, version=version, applies_to=self.applies_to)
 
     def __str__(self):
         """
